# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect
from django.views.generic.base import TemplateResponseMixin, View
from django.views.generic.edit import FormMixin

from .. import forms
from .detail import BaseDetailView, SingleObjectMixin, SingleObjectTemplateResponseMixin


class ModelFormMixin(FormMixin, SingleObjectMixin):
    fields = None

    def get_form_class(self):
        if self.fields is not None and self.form_class:
            raise ImproperlyConfigured("Specifying both 'fields' and 'form_class' is not permitted.")

        if self.form_class:
            return self.form_class

        model = self.get_model()
        return forms.modelform_factory(model, fields=self.fields, session=self.session)

    def get_form_kwargs(self):
        """
        Return the keyword arguments for instantiating the form.
        """
        kwargs = super(ModelFormMixin, self).get_form_kwargs()
        if hasattr(self, "object"):
            kwargs.update({"instance": self.object})

        kwargs["session"] = self.session
        return kwargs

    def get_success_url(self):
        """
        Return the URL to redirect to after processing a valid form.
        """
        if self.success_url:
            return self.success_url.format(**vars(self.object))

        raise ImproperlyConfigured(
            "No URL to redirect to. Either provide a url or override this function to return a url"
        )

    def form_valid(self, form):
        self.object = form.save()
        return super(ModelFormMixin, self).form_valid(form)


class ProcessFormView(View):
    """Render a form on GET and processes it on POST."""

    def get(self, request, *args, **kwargs):
        """Handle GET requests: instantiate a blank version of the form."""
        return self.render_to_response(self.get_context_data())

    def post(self, request, *args, **kwargs):
        """
        Handle POST requests: instantiate a form instance with the passed
        POST variables and then check if it's valid.
        """
        form = self.get_form()
        if form.is_valid():
            return self.form_valid(form)

        return self.form_invalid(form)

    put = post


class BaseFormView(FormMixin, ProcessFormView):
    """A base view for displaying a form."""


class FormView(TemplateResponseMixin, BaseFormView):
    """A view for displaying a form and rendering a template response."""


class BaseCreateView(ModelFormMixin, ProcessFormView):
    """
    Base view for creating a new object instance.
    Using this base class requires subclassing to provide a response mixin.
    """

    def get(self, request, *args, **kwargs):
        self.object = None
        return super(BaseCreateView, self).get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object = None
        return super(BaseCreateView, self).post(request, *args, **kwargs)


class CreateView(SingleObjectTemplateResponseMixin, BaseCreateView):
    """
    View for creating a new object, with a response rendered by a template.
    """

    template_name_suffix = "_form"


class BaseUpdateView(ModelFormMixin, ProcessFormView):
    """
    Base view for updating an existing object.
    Using this base class requires subclassing to provide a response mixin.
    """

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super(BaseUpdateView, self).get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super(BaseUpdateView, self).post(request, *args, **kwargs)


class UpdateView(SingleObjectTemplateResponseMixin, BaseUpdateView):
    """View for updating an object, with a response rendered by a template."""

    template_name_suffix = "_form"


class DeletionMixin:
    """Provide the ability to delete objects."""

    success_url = None

    def delete(self, request, *args, **kwargs):
        """
        Call the delete() method on the fetched object and then redirect to the
        success URL.
        """
        self.object = self.get_object()
        success_url = self.get_success_url()
        self.session.delete(self.object)
        self.session.flush()
        return HttpResponseRedirect(success_url)

    # Add support for browsers which only accept GET and POST for now.

    def post(self, request, *args, **kwargs):
        return self.delete(request, *args, **kwargs)

    def get_success_url(self):
        if self.success_url:
            return self.success_url.format(**self.object.__dict__)

        else:
            raise ImproperlyConfigured("No URL to redirect to. Provide a success_url.")


class BaseDeleteView(DeletionMixin, BaseDetailView):
    """
    Base view for deleting an object.
    Using this base class requires subclassing to provide a response mixin.
    """


class DeleteView(SingleObjectTemplateResponseMixin, BaseDeleteView):
    """
    View for deleting an object retrieved with self.get_object(), with a
    response rendered by a template.
    """

    template_name_suffix = "_confirm_delete"



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
import os

import sqlalchemy as sa

import django

from django_sorcery.db import signals


os.environ["DJANGO_SETTINGS_MODULE"] = "tests.settings"


@signals.engine_created.connect
def hacks(sender, **kwargs):

    # pysqlite workarounds for savepoints
    if "sqlite" in sender.url.drivername:

        @sa.event.listens_for(sender, "connect")
        def do_connect(dbapi_connection, connection_record):
            # disable pysqlite's emitting of the BEGIN statement entirely.
            # also stops it from emitting COMMIT before any DDL.
            dbapi_connection.isolation_level = None

        @sa.event.listens_for(sender, "begin")
        def do_begin(conn):
            # emit our own BEGIN
            conn.execute("BEGIN")


django.setup()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import sqlalchemy as sa

from django_sorcery.db import meta  # noqa

from ..base import TestCase
from ..models import COLORS, AllKindsOfFields, Owner, Vehicle, VehicleType, Vertex


class TestModelMeta(TestCase):
    def test_model_meta(self):
        info = meta.model_info(Owner)

        self.assertEqual(repr(info), "<model_info(Owner)>")
        self.assertEqual(set(info.primary_keys.keys()), {"id"})
        self.assertEqual(info.primary_keys["id"].property, Owner.id.property)

        self.assertEqual(set(info.properties.keys()), {"first_name", "last_name"})
        self.assertEqual(info.properties["first_name"].property, Owner.first_name.property)
        self.assertEqual(info.properties["last_name"].property, Owner.last_name.property)

        self.assertEqual(set(info.relationships.keys()), {"vehicles"})
        self.assertTrue(info.relationships["vehicles"].relationship is Owner.vehicles.property)

        self.assertEqual(set(info.field_names), {"id", "first_name", "last_name", "vehicles"})

    def test_model_meta_with_mapper(self):
        mapper = Vehicle.owner.property.parent
        self.assertEqual(meta.model_info(mapper), meta.model_info(Vehicle))


class TestCompositeMeta(TestCase):
    def test_composite_meta(self):

        info = meta.model_info(Vertex)
        self.assertEqual(set(info.composites.keys()), {"start", "end"})

        start = info.composites["start"]
        self.assertEqual(set(start.properties.keys()), {"x", "y"})
        self.assertEqual(start.properties["x"].property, Vertex.x1.property)

        self.assertEqual(set(start.field_names), {"x", "y"})


class TestRelationshipMeta(TestCase):
    def test_relationship_meta(self):
        info = meta.model_info(Owner)

        rel = info.relationships["vehicles"]

        self.assertEqual(rel.related_model, Vehicle)
        self.assertEqual(rel.name, "vehicles")
        self.assertEqual(rel.direction, sa.orm.relationships.ONETOMANY)
        self.assertEqual(list(rel.foreign_keys), Vehicle._owner_id.property.columns)
        self.assertTrue(rel.uselist)
        self.assertEqual(repr(rel), "<relation_info(Owner.vehicles)>")


class TestColumnMeta(TestCase):
    def test_columnInfo(self):
        info = meta.model_info(Vehicle)

        col = info.primary_keys["id"]
        self.assertEqual(col.name, "id")
        self.assertDictEqual({"label": "Id", "help_text": "The primary key", "required": True}, col.field_kwargs)

        col = info.properties["type"]
        self.assertDictEqual(
            {"enum_class": VehicleType, "help_text": None, "label": "Type", "max_length": 3, "required": True},
            col.field_kwargs,
        )

        col = info.properties["paint"]
        self.assertDictEqual(
            {"choices": COLORS, "help_text": None, "label": "Paint", "max_length": 6, "required": False},
            col.field_kwargs,
        )
        self.assertEqual(col.parent_model, Vehicle)
        self.assertEqual(repr(col), "<column_info(Vehicle.paint)>")

        info = meta.model_info(AllKindsOfFields)
        col = info.properties["decimal"]
        self.assertDictEqual(
            {"decimal_places": None, "help_text": None, "label": "Decimal", "max_digits": None, "required": False},
            col.field_kwargs,
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django_sorcery.db.query import QueryProperty

from ..base import TestCase
from ..models import CompositePkModel, Owner, Vehicle, VehicleType, db


class TestQuery(TestCase):
    def setUp(self):
        super(TestQuery, self).setUp()
        db.add_all(
            [
                Owner(id=1, first_name="Test 1", last_name="Owner 1"),
                Owner(id=2, first_name="Test 2", last_name="Owner 2"),
                Owner(id=3, first_name="Test 3", last_name="Owner 3"),
                Owner(id=4, first_name="Test 4", last_name="Owner 4"),
            ]
        )
        db.add(CompositePkModel(id=1, pk=1, active=True, name="Test-1-1"))
        db.flush()

    def tearDown(self):
        super(TestQuery, self).tearDown()
        db.rollback()
        db.remove()

    def test_query_get_regular(self):
        owner = Owner.query.get(1)
        self.assertEqual(owner.id, 1)

    def test_query_get_with_kwargs(self):
        owner = Owner.query.get(id=1)
        self.assertEqual(owner.id, 1)

    def test_query_get_with_kwargs_composite(self):
        obj = CompositePkModel.query.get(id=1, pk=1)
        self.assertEqual(obj.name, "Test-1-1")

    def test_query_get_with_kwargs_tuple(self):
        obj = CompositePkModel.query.get((1, 1))
        self.assertEqual(obj.name, "Test-1-1")


class TestQueryProperty(TestCase):
    def setUp(self):
        super(TestQueryProperty, self).setUp()
        db.add_all(
            [
                Vehicle(name="used", is_used=True, type=VehicleType.car),
                Vehicle(name="new", is_used=False, type=VehicleType.car),
            ]
        )
        db.flush()

    def test_options_filter_and_filter_by(self):
        class Dummy(object):
            vehicles = QueryProperty(db, Vehicle).options(db.joinedload(Vehicle.owner))
            used_vehicles = QueryProperty(db, Vehicle, Vehicle.is_used.is_(True))
            new_vehicles = QueryProperty(db, Vehicle, is_used=False)

        dummy = Dummy()

        self.assertEqual(dummy.vehicles.count(), 2)
        self.assertEqual(dummy.used_vehicles.count(), 1)
        self.assertEqual(dummy.new_vehicles.count(), 1)

    def test_bad_attr(self):

        with self.assertRaises(AttributeError) as ctx:
            QueryProperty(db, Vehicle).dummy(db.joinedload(Vehicle.owner))

        self.assertEqual(
            ctx.exception.args,
            ("<QueryProperty db=<SQLAlchemy engine=sqlite://>, model='Vehicle'> object has no attribute 'dummy'",),
        )

    def test_no_model(self):
        class Dummy(object):
            vehicles = QueryProperty(db, None)

        with self.assertRaises(AttributeError) as ctx:
            Dummy.vehicles

        self.assertEqual(
            ctx.exception.args,
            (
                "Cannot access QueryProperty when not bound to a model. You can explicitly instantiate descriptor with "
                "model class - `db.queryproperty(Model)`.",
            ),
        )

    def test_stupid(self):

        qp = QueryProperty(db, Vehicle)

        class Dummy(object):
            vehicles = qp

        qp.model = object

        self.assertIsNone(Dummy.vehicles)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django import forms as djangoforms
from django.forms import fields as djangofields

from django_sorcery import fields
from django_sorcery.db.meta import model_info
from django_sorcery.field_mapping import ModelFieldMapper
from django_sorcery.forms import SQLAModelFormOptions

from .base import TestCase
from .models import AllKindsOfFields, Owner, Vehicle, db


class TestFieldMapping(TestCase):
    def test_get_default_kwargs(self):
        class Meta:
            model = Owner
            session = db.session
            widgets = {"test": "widget"}
            labels = {"test": "label"}
            help_texts = {"test": "help_text"}
            error_messages = {"test": "error_message"}
            field_classes = {"test": "field_class"}
            localized_fields = ("test",)

        opts = SQLAModelFormOptions(Meta)
        mapper = ModelFieldMapper(opts)

        self.assertDictEqual({}, mapper.get_default_kwargs("none"))

        self.assertDictEqual(
            {
                "error_messages": "error_message",
                "form_class": "field_class",
                "help_text": "help_text",
                "label": "label",
                "localize": True,
                "widget": "widget",
            },
            mapper.get_default_kwargs("test"),
        )

        mapper.opts.localized_fields = "__all__"
        self.assertDictEqual(
            {
                "error_messages": "error_message",
                "form_class": "field_class",
                "help_text": "help_text",
                "label": "label",
                "localize": True,
                "widget": "widget",
            },
            mapper.get_default_kwargs("test"),
        )

    def test_get_fields(self):
        class Meta:
            model = Owner
            session = db

        mapper = ModelFieldMapper(SQLAModelFormOptions(Meta))

        fields = mapper.get_fields()

        self.assertListEqual(sorted(list(fields.keys())), ["first_name", "last_name", "vehicles"])

    def test_get_fields_with_include(self):
        class Meta:
            model = Owner
            fields = ["first_name"]
            session = db

        mapper = ModelFieldMapper(SQLAModelFormOptions(Meta))

        fields = mapper.get_fields()

        self.assertListEqual(list(fields.keys()), ["first_name"])

    def test_get_fields_with_exclude(self):
        class Meta:
            model = Owner
            exclude = ["first_name", "vehicles"]
            session = db

        mapper = ModelFieldMapper(SQLAModelFormOptions(Meta))

        fields = mapper.get_fields()

        self.assertListEqual(list(fields.keys()), ["last_name"])

    def test_get_fields_formfield_callback(self):

        info = model_info(Owner)
        self.called_with = []

        def callback(attr, **kwargs):
            self.called_with.append(attr)
            return djangofields.CharField()

        class Meta:
            model = Owner
            fields = ["first_name", "vehicles"]

        mapper = ModelFieldMapper(SQLAModelFormOptions(Meta), formfield_callback=callback)

        fields = mapper.get_fields()

        self.assertListEqual(list(fields.keys()), ["first_name", "vehicles"])
        self.assertEqual(self.called_with, [info.properties["first_name"], info.relationships["vehicles"]])

    def test_get_fields_no_private(self):
        class Meta:
            model = Vehicle
            session = db

        mapper = ModelFieldMapper(SQLAModelFormOptions(Meta))

        fields = mapper.get_fields()

        self.assertEqual(
            list(sorted(fields.keys())), ["created_at", "is_used", "name", "options", "owner", "paint", "parts", "type"]
        )
        self.assertNotIn(Vehicle._owner_id.key, fields)


class TestTypeMap(TestCase):
    def setUp(self):
        super(TestTypeMap, self).setUp()

        class Meta:
            model = AllKindsOfFields
            session = db

        self.mapper = ModelFieldMapper(SQLAModelFormOptions(Meta))
        self.fields = self.mapper.get_fields()

    def test_boolean_field_not_nullable(self):
        self.assertIn("boolean_notnull", self.fields)
        field = self.fields["boolean_notnull"]
        self.assertIsInstance(field, djangofields.BooleanField)

    def test_boolean_field(self):
        self.assertIn("boolean", self.fields)
        field = self.fields["boolean"]
        self.assertIsInstance(field, djangofields.NullBooleanField)

    def test_enum_field(self):
        self.assertIn("enum", self.fields)
        field = self.fields["enum"]
        self.assertIsInstance(field, fields.EnumField)

    def test_enum_choice_field(self):
        self.assertIn("enum_choice", self.fields)
        field = self.fields["enum_choice"]
        self.assertIsInstance(field, djangofields.TypedChoiceField)

    def test_bigint_field(self):
        self.assertIn("bigint", self.fields)
        field = self.fields["bigint"]
        self.assertIsInstance(field, djangofields.IntegerField)

    def test_biginteger_field(self):
        self.assertIn("biginteger", self.fields)
        field = self.fields["biginteger"]
        self.assertIsInstance(field, djangofields.IntegerField)

    def test_decimal_field(self):
        self.assertIn("decimal", self.fields)
        field = self.fields["decimal"]
        self.assertIsInstance(field, djangofields.DecimalField)

    def test_float_field(self):
        self.assertIn("float", self.fields)
        field = self.fields["float"]
        self.assertIsInstance(field, djangofields.FloatField)

    def test_int_field(self):
        self.assertIn("int", self.fields)
        field = self.fields["int"]
        self.assertIsInstance(field, djangofields.IntegerField)

    def test_integer_field(self):
        self.assertIn("integer", self.fields)
        field = self.fields["integer"]
        self.assertIsInstance(field, djangofields.IntegerField)

    def test_numeric_field(self):
        self.assertIn("numeric", self.fields)
        field = self.fields["numeric"]
        self.assertIsInstance(field, djangofields.DecimalField)

    def test_real_field(self):
        self.assertIn("real", self.fields)
        field = self.fields["real"]
        self.assertIsInstance(field, djangofields.FloatField)

    def test_smallint_field(self):
        self.assertIn("smallint", self.fields)
        field = self.fields["smallint"]
        self.assertIsInstance(field, djangofields.IntegerField)

    def test_smallinteger_field(self):
        self.assertIn("smallinteger", self.fields)
        field = self.fields["smallinteger"]
        self.assertIsInstance(field, djangofields.IntegerField)

    def test_char_field(self):
        self.assertIn("char", self.fields)
        field = self.fields["char"]
        self.assertIsInstance(field, djangofields.CharField)

    def test_clob_field(self):
        self.assertIn("clob", self.fields)
        field = self.fields["clob"]
        self.assertIsInstance(field, djangofields.CharField)
        self.assertIsInstance(field.widget, djangoforms.Textarea)

    def test_nchar_field(self):
        self.assertIn("nchar", self.fields)
        field = self.fields["nchar"]
        self.assertIsInstance(field, djangofields.CharField)

    def test_nvarchar_field(self):
        self.assertIn("nvarchar", self.fields)
        field = self.fields["nvarchar"]
        self.assertIsInstance(field, djangofields.CharField)

    def test_string_field(self):
        self.assertIn("string", self.fields)
        field = self.fields["string"]
        self.assertIsInstance(field, djangofields.CharField)

    def test_text_field(self):
        self.assertIn("text", self.fields)
        field = self.fields["text"]
        self.assertIsInstance(field, djangofields.CharField)
        self.assertIsInstance(field.widget, djangoforms.Textarea)

    def test_unicode_field(self):
        self.assertIn("unicode", self.fields)
        field = self.fields["unicode"]
        self.assertIsInstance(field, djangofields.CharField)

    def test_unicodetext_field(self):
        self.assertIn("unicodetext", self.fields)
        field = self.fields["unicodetext"]
        self.assertIsInstance(field, djangofields.CharField)
        self.assertIsInstance(field.widget, djangoforms.Textarea)

    def test_varchar_field(self):
        self.assertIn("varchar", self.fields)
        field = self.fields["varchar"]
        self.assertIsInstance(field, djangofields.CharField)

    def test_date_field(self):
        self.assertIn("date", self.fields)
        field = self.fields["date"]
        self.assertIsInstance(field, djangofields.DateField)

    def test_datetime_field(self):
        self.assertIn("datetime", self.fields)
        field = self.fields["datetime"]
        self.assertIsInstance(field, djangofields.DateTimeField)

    def test_interval_field(self):
        self.assertIn("interval", self.fields)
        field = self.fields["interval"]
        self.assertIsInstance(field, djangofields.DurationField)

    def test_time_field(self):
        self.assertIn("time", self.fields)
        field = self.fields["time"]
        self.assertIsInstance(field, djangofields.TimeField)

    def test_timestamp_field(self):
        self.assertIn("timestamp", self.fields)
        field = self.fields["timestamp"]
        self.assertIsInstance(field, djangofields.DateTimeField)

    def test_binary_field(self):
        self.assertNotIn("binary", self.fields)

    def test_blob_field(self):
        self.assertNotIn("blob", self.fields)

    def test_largebinary_field(self):
        self.assertNotIn("largebinary", self.fields)

    def test_varbinary_field(self):
        self.assertNotIn("varbinary", self.fields)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


class ValidateTogetherModelFields(object):
    """
    Validator for checking that multiple model fields are always saved together

    For example::

        class MyModel(db.Model):
            foo = db.Column(db.Integer())
            bar = db.Column(db.Integer())

            validators = [
                ValidateTogetherModelFields(["foo", "bar"]),
            ]
    """

    message = _("All %(fields)s are required.")
    code = "required"

    def __init__(self, fields, message=None, code=None):
        self.fields = fields
        self.message = message or self.message
        self.code = code or self.code

    def __call__(self, m):
        if not any(getattr(m, i, None) for i in self.fields):
            return
        if not all(getattr(m, i, None) for i in self.fields):
            raise ValidationError(self.message, code=self.code, params={"fields": ", ".join(sorted(self.fields))})



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django.db import DEFAULT_DB_ALIAS
from django.utils.module_loading import import_string

from ..utils import suppress
from .sqlalchemy import SQLAlchemy
from .url import get_settings


class dbdict(dict):
    def get(self, alias=None, cls=SQLAlchemy, **kwargs):
        """
        Returns a :py:class:`..sqlalchemy.SQLAlchemy` instance from configuration and registers it. Can return a custom
        :py:class:`..sqlalchemy.SQLAlchemy` instance thru args or thru ``SQLALCHEMY`` database setting in configuration.
        """
        alias = alias or DEFAULT_DB_ALIAS
        if alias in self:
            return self[alias]

        with suppress(Exception):
            settings = get_settings(alias)
            cls = import_string(settings.get("SQLALCHEMY"))

        assert SQLAlchemy in cls.mro(), "'%s' needs to subclass from SQLAlchemy" % cls.__name__

        return self.setdefault(alias, cls(alias=alias, **kwargs))

    def update(self, *args, **kwargs):
        for arg in args:
            other = dict(arg)
            for key in other:
                self[key] = other[key]
        for key in kwargs:
            self[key] = kwargs[key]

    def __setitem__(self, alias, val):
        if alias in self:
            raise RuntimeError("Database alias `{alias}` has already been created".format(alias=alias))

        if val in self.values():
            raise RuntimeError("Database alias `{alias}` has already been created".format(alias=alias))

        if not isinstance(val, SQLAlchemy):
            raise RuntimeError("Database alias `{alias}` has wrong type".format(alias=alias))

        super(dbdict, self).__setitem__(alias, val)

    def rollback(self):
        """
        Applies rollback on all registered databases
        """
        for db in self.values():
            db.rollback()

    def flush(self):
        """
        Applies flush on all registered databases
        """
        for db in self.values():
            db.flush()

    def commit(self):
        """
        Applies commit on all registered databases
        """
        for db in self.values():
            db.commit()

    def remove(self):
        """
        Applies remove on all registered databases
        """
        for db in self.values():
            db.remove()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from collections import namedtuple

from django.core.exceptions import ValidationError
from django.test import TestCase

from django_sorcery.validators import ValidateTogetherModelFields


Node = namedtuple("Node", ["left", "right", "value"])
Node.__new__.__defaults__ = (None, None, None)


class TestValidateTogetherModelFields(TestCase):
    def test_success_all_none(self):
        node = Node()

        validator = ValidateTogetherModelFields(["left", "right", "value"])

        self.assertIsNone(validator(node))

    def test_success_all_full(self):
        node = Node("left", "right", "value")

        validator = ValidateTogetherModelFields(["left", "right", "value"])

        self.assertIsNone(validator(node))

    def test_success_fail(self):
        node = Node("left", "right")

        validator = ValidateTogetherModelFields(["left", "right", "value"])

        with self.assertRaises(ValidationError):
            self.assertIsNone(validator(node))
